/*
 * Simple "bit-banging" test application for the athernet's CPLD SPI-Port
 * Demonstrates the access to banked SRAM
 *
 * 2005 by Patrick 'OldBug' Dohmen
 */

#include <avr/io.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "extmem.h"
#include "cpld.h"

/* Prototypes: */
void ioinit(void);
void bitbangCPLDSPI_CSTest(void);
void usartinit(void);
void uart_puts (char *);
void uart_putc(unsigned char);
void mem_dump(void *, size_t);

/* Local defines: */
#define nop()	asm volatile( "nop\n\t":: )

/* USART-Init beim ATmegaXX */
#define F_OSC           14745600L           /* Oszillator-Frequenz in Hz */
#define UART_BAUD_RATE  115200L
#define UART_BAUD_CALC(UART_BAUD_RATE, F_OSC) ((F_OSC) / ((UART_BAUD_RATE) * 16L ) - 1L)

/* Global variables: */

int
main(void)
{
	volatile unsigned char *pSRAM = NULL;
	uint16_t tStart = 0, tStop = 0;
	unsigned char outbuf[128];
	uint8_t data = 0;
	uint8_t *pExtMallocBlock = NULL;
	
	/* init hardware */
	ioinit();
	usartinit();

	/* run forever */
	while(1)
	{
		PORTB &= ~(1 << PB5);

		sprintf(outbuf, "Heap-Start: %p; -End: %p\r\n", __malloc_heap_start, __malloc_heap_end);
		uart_puts(outbuf);

#		if 1
		sprintf(outbuf, "\r\n-- Starting external SRAM test --\r\n");
		uart_puts(outbuf);

		sprintf(outbuf, "   o writing %d bytes...\r\n", ((uint8_t *) 0xFFFF) - ((uint8_t *) 0x8000));
		uart_puts(outbuf);

		data = 0;

#		if 1
		for(pSRAM = pBank_1; pSRAM < (unsigned char *) 0xFFFF; pSRAM++)
		{
			*pSRAM = data++;
		}
#		endif

		TCNT3 = 0x0000;
		TCCR3B = (1 << CS30);

		tStart = TCNT3;
		pExtMallocBlock = malloc(256);
		tStop = TCNT3;

		memset(pExtMallocBlock, 0xAA, 256);
		
		sprintf(outbuf, "   o comparing...\r\n");
		uart_puts(outbuf);

		data = 0;

#		if 1
		for(pSRAM = pBank_1; pSRAM < (unsigned char *) 0xFFFF; pSRAM++)
		{
			if(*pSRAM != data)
			{
				sprintf(outbuf, "\r\n   ** @0x%p ERROR; expected: 0x%02X, read: 0x%02X\r\n", pSRAM, data, *pSRAM);
				uart_puts(outbuf);
			}
			data++;
		}
#		endif

		sprintf(outbuf, "-- Test done, it took %d cycles. --\r\n", (tStop - tStart));
		uart_puts(outbuf);

		PORTB |= (1 << PB5);

		sprintf(outbuf, "-- Dump memory --\r\n\r\n");
		uart_puts(outbuf);

		mem_dump(pExtMallocBlock, (size_t) 0xFF);
		free(pExtMallocBlock);
		
		mem_dump((void *)0x8000, (size_t) 0xFF /*((uint16_t) 0xFFFF - (uint16_t) 0x8000)*/);
		
		sprintf(outbuf, "-- End of memory dump --\r\n");
		uart_puts(outbuf);
#		endif
		
		/* do some bit banging on the SPI-Port */
		bitbangCPLDSPI_CSTest();
	}

	/* NEVEREACHED */
	return 0;
}

void
ioinit(void)
{
	DDRB = (1 << PB5) | (1 << PB4);		/* PB5 as output (LED)		*/
	PORTB |= (1 << PB5);			/* LED off			*/
}

void
bitbangCPLDSPI_CSTest(void)
{
	/* RMW (toggle pin) */
	*pSPI_CS ^= (1 << SPIP0);
}

void
usartinit(void)
{
        UBRR1H = (unsigned char) (UART_BAUD_CALC(UART_BAUD_RATE, F_OSC) >> 8);
        UBRR1L = (unsigned char) UART_BAUD_CALC(UART_BAUD_RATE, F_OSC);
        UCSR1B = (1 << TXEN);
        UCSR1C = (1 << UCSZ01) | (1 << UCSZ00);
	UDR0 = 0x00;
}

void
uart_putc(unsigned char c)
{
        while(!(UCSR1A & (1 << UDRE1)))
                ;

        UDR1 = c;
}

void
uart_puts(char *s)
{
        while (*s)
        {
                uart_putc(*s);
                s++;
        }
}

void
mem_dump(void *adr, size_t len)
{
	unsigned char outbuf[64];
	unsigned char c[17];
	uint8_t *start = adr;
	unsigned char i;

#	if 0
	sprintf(outbuf, "Dumping %d bytes:\r\n\r\n", len);
	uart_puts(outbuf);
#	endif
	
	sprintf(outbuf, "%p: 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\r\n", adr);
	uart_puts(outbuf);
		
	for(; ((uint8_t *) adr) < ((uint8_t *)(((uint16_t) start) + ((uint16_t) len))); adr += 16)
	{
		sprintf(outbuf, "%p: ", adr);
		uart_puts(outbuf);
		
		for(i = 0; i < 16; i++)
		{
			c[i] = (*((unsigned char *) (adr + i)) > 0x20) ? (*((unsigned char *) (adr + i))) : ('.');

#			if 1
			sprintf(outbuf, "%02X ", *((uint8_t *)(adr + i)));	/* does not work with minimalistic printf-variant */
#			endif

#			if 0			
			/* Workaround for minimalistic printf-variant */
			if((*((uint8_t *)(adr + i))) >= 0x10)
				sprintf(outbuf, "%X ", *((uint8_t *)(adr + i)));
			else
				sprintf(outbuf, "0%X ", *((uint8_t *)(adr + i)));
#			endif
				
			uart_puts(outbuf);
		}

		c[16] = 0;

		sprintf(outbuf, ":%s:\r\n", c);
		uart_puts(outbuf);
	}
}
